import numpy as np
import torch
import torch.jit
from sklearn.compose import TransformedTargetRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVR

from .ttr import TorchTransformedTargetRegressor


def test_ttr():
    x = np.random.random(size=(100, 2)) * 2 - 1
    y = 10000 * (np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2) - 20000

    sk_obj = TransformedTargetRegressor(regressor=SVR(), transformer=StandardScaler())
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchTransformedTargetRegressor.wrap(sk_obj))

    with torch.no_grad():
        x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert (np.abs(expected - actual) < 0.1).all()



from typing import List, Type

import numpy as np
import torch
import torch.jit
import torch.nn as nn
from sklearn.linear_model import LinearRegression


class TorchLinearRegression(nn.Module):
    def __init__(
        self,
        weights: torch.Tensor,
        biases: torch.Tensor,
    ):
        super().__init__()
        if len(weights.shape) == 1:
            weights = weights[None]
        self.weights = nn.Parameter(weights)
        self.biases = nn.Parameter(biases)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [LinearRegression]

    @classmethod
    def wrap(cls, obj: LinearRegression) -> "TorchLinearRegression":
        return cls(
            weights=torch.from_numpy(obj.coef_),
            biases=torch.from_numpy(np.array(obj.intercept_)),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.

        An alias for self.predict().
        """
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.
        """
        outputs = (x @ self.weights.t()) + self.biases
        if outputs.shape[1] == 1:
            return outputs.view(-1)
        return outputs



import itertools
from collections import Counter

import numpy as np
import pytest
import torch
from sklearn.dummy import DummyClassifier

from .dummy import TorchDummyClassifier


@pytest.mark.parametrize(
    ("strategy", "multi_class"),
    list(
        itertools.product(
            ["most_frequent", "prior", "stratified", "uniform", "constant"],
            [False, True],
        )
    ),
)
def test_dummy_classifier(strategy: str, multi_class: bool):
    rs = np.random.RandomState(1337)
    xs = rs.normal(size=(10000, 3))
    if multi_class:
        ys = np.concatenate(
            [
                rs.randint(low=1, high=3, size=(xs.shape[0], 1)),
                rs.randint(low=-1, high=5, size=(xs.shape[0], 1)),
            ],
            axis=1,
        )
        constant = np.array([2, 3])
    else:
        ys = rs.choice(
            a=[-2, 1, 3, 5, 7, 6],
            p=[1 / 6, 1 / 6 - 1 / 12, 1 / 6 + 1 / 12, 1 / 6, 1 / 6, 1 / 6],
            size=(xs.shape[0],),
        )
        constant = 1

    sk_obj = DummyClassifier(
        strategy=strategy, constant=constant if strategy == "constant" else None
    )
    sk_obj.fit(xs, ys)
    th_obj = TorchDummyClassifier.wrap(sk_obj)

    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        expected = sk_obj.predict(xs)
        actual = th_obj.predict(xs_th).numpy()
        assert expected.shape == actual.shape
        if strategy in ["most_frequent", "prior", "constant"]:
            assert (actual == expected).all()
        else:
            # Make sure the sampling probabilities are the same, even though
            # the outputs are non-deterministic.
            threshold = 0.05
            if len(expected.shape) == 1:
                actual, expected = actual[:, None], expected[:, None]
            for i in range(expected.shape[1]):
                x, a = expected[:, i], actual[:, i]
                assert x.shape == a.shape
                x_counts = Counter(x)
                a_counts = Counter(a)
                assert set(x_counts.keys()) == set(a_counts.keys())
                for k, x_count in x_counts.items():
                    a_count = a_counts[k]
                    assert abs(x_count / len(x) - a_count / len(a)) < threshold

        expected = sk_obj.predict_proba(xs)
        actual = th_obj.predict_proba(xs_th)
        assert len(expected) == len(actual)
        if strategy in ["most_frequent", "prior", "uniform", "constant"]:
            for x, a in zip(expected, actual):
                assert x.shape == a.shape
                assert np.allclose(x, a.numpy())
        else:
            # Make sure the sampling probabilities are the same, even though
            # the outputs are non-deterministic.
            threshold = 0.05
            if not isinstance(expected, list):
                expected, actual = [expected], [actual]
            for x, a in zip(expected, actual):
                a = a.numpy()
                assert not np.allclose(x, a)
                assert x.shape == a.shape
                assert (
                    np.max(np.abs(np.mean(x, axis=0) - np.mean(a, axis=0))) < threshold
                )



from typing import Callable, Tuple

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.datasets import load_breast_cancer, load_digits
from sklearn.linear_model import SGDClassifier

from .sgd_classifier import TorchSGDClassifier


def xor_dataset(**_) -> Tuple[np.ndarray, np.ndarray]:
    """
    A binary classification problem that is not linearly separable.
    """
    x = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    y = np.array([0.0, 1.0, 1.0, 0.0])
    return np.tile(x, [10, 1]), np.tile(y, [10])


def xor_and_dataset(**_) -> Tuple[np.ndarray, np.ndarray]:
    """
    A three-class problem that is not linearly separable.
    """
    x = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    y = np.array([0, 1, 1, 2])
    return np.tile(x, [10, 1]), np.tile(y, [10])


@pytest.mark.parametrize(
    ("dataset", "loss", "check_probs", "fit_intercept", "space_classes"),
    [
        (load_breast_cancer, "log", False, True, False),
        (load_breast_cancer, "log", False, True, False),
        (load_digits, "log", False, True, False),
        (load_digits, "log", False, True, True),
        (xor_dataset, "log", True, True, False),
        (xor_and_dataset, "log", True, True, False),
        (xor_and_dataset, "log", True, False, False),
        (xor_and_dataset, "hinge", False, True, False),
    ],
)
def test_sgd_classifier(
    dataset: Callable[..., Tuple[np.ndarray, np.ndarray]],
    loss: str,
    check_probs: bool,
    fit_intercept: bool,
    space_classes: bool,
):
    x, y = dataset(return_X_y=True)
    if space_classes:
        n_classes = np.max(y)
        y = np.where(y > n_classes // 2, 1 + (n_classes - y) + n_classes // 2, y)

    sk_obj = SGDClassifier(loss=loss, fit_intercept=fit_intercept)
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchSGDClassifier.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.decision_function(x)
        actual = th_obj.decision_function(x_th).numpy()
        assert np.allclose(expected, actual)

        expected = sk_obj.predict(x)
        actual = th_obj(x_th).numpy()
        assert (expected == actual).all()

        if check_probs:
            expected = sk_obj.predict_log_proba(x)
            actual = th_obj.predict_log_proba(x_th).numpy()
            assert not np.isnan(expected).any()
            assert np.allclose(np.exp(expected), np.exp(actual))

            expected = sk_obj.predict_proba(x)
            actual = th_obj.predict_proba(x_th).numpy()
            assert np.isfinite(expected).any()
            assert np.isfinite(actual).any()
            assert np.allclose(expected, actual)



from typing import List, Type, Union

import numpy as np
import torch
import torch.jit
import torch.nn as nn
from sklearn.svm import SVR, LinearSVR, NuSVR

from .kernel import Kernel


class TorchSVR(nn.Module):
    def __init__(
        self,
        kernel: Kernel,
        support_vectors: torch.Tensor,
        dual_coef: torch.Tensor,
        intercept: torch.Tensor,
    ):
        super().__init__()
        self.kernel = kernel
        self.support_vectors = nn.Parameter(support_vectors)
        self.dual_coef = nn.Parameter(dual_coef)
        self.intercept = nn.Parameter(intercept)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [SVR, NuSVR]

    @classmethod
    def wrap(cls, obj: Union[SVR, NuSVR]) -> "TorchSVR":
        assert not obj._sparse, "sparse SVR not supported"
        return cls(
            kernel=Kernel.wrap(obj),
            support_vectors=torch.from_numpy(obj.support_vectors_),
            dual_coef=torch.from_numpy(obj.dual_coef_),
            intercept=torch.from_numpy(obj.intercept_),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict regression values for the given feature vectors.

        An alias for self.predict().
        """
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.
        """
        kernel_out = self.kernel(x, self.support_vectors)
        return torch.einsum("jk,jk->j", self.dual_coef, kernel_out) + self.intercept


class TorchLinearSVR(nn.Module):
    def __init__(self, coef: torch.Tensor, intercept: torch.Tensor):
        super().__init__()
        self.coef = nn.Parameter(coef)
        self.intercept = nn.Parameter(intercept)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [LinearSVR]

    @classmethod
    def wrap(cls, obj: LinearSVR) -> "TorchLinearSVR":
        assert isinstance(obj.coef_, np.ndarray), "sparse LinearSVR is not supported"
        coef = torch.from_numpy(obj.coef_)
        return cls(
            coef=coef,
            intercept=(
                torch.from_numpy(obj.intercept_)
                if obj.fit_intercept
                else torch.zeros_like(coef[0])
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        return x @ self.coef + self.intercept



import numpy as np
import pytest
import torch
from sklearn.linear_model import LinearRegression

from .linear_regression import TorchLinearRegression


@pytest.mark.parametrize(
    (
        "num_targets",
        "fit_intercept",
    ),
    [
        (1, False),
        (1, True),
        (2, False),
        (2, True),
    ],
)
def test_linear_svr(num_targets: int, fit_intercept: bool):
    x = np.random.random(size=(100, 2)) * 2 - 1
    if num_targets == 1:
        y = np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2
    else:
        assert num_targets == 2
        y = np.stack(
            [
                np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2,
                np.cos(x[:, 0]) + np.sin(x[:, 1]) * 0.2,
            ],
            axis=1,
        )

    sk_obj = LinearRegression(fit_intercept=fit_intercept)
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchLinearRegression.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert actual.shape == expected.shape
        assert (np.abs(expected - actual) < 1e-5).all()



import numpy as np
import pytest
import torch
import torch.jit
from sklearn.preprocessing import StandardScaler

from .standard_scaler import TorchStandardScaler


@pytest.mark.parametrize(
    ("with_mean", "with_std"), [(True, True), (True, False), (False, True)]
)
def test_standard_scaler(with_mean, with_std):
    x = np.random.normal(size=(30, 3))
    x *= np.array([2.0, 0.3, 1.0])
    x += np.array([1.0, 2.0, 3.0])
    sk_obj = StandardScaler(with_mean=with_mean, with_std=with_std)
    sk_obj.fit(x)
    th_obj = torch.jit.script(TorchStandardScaler.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.transform(x, copy=True)
        actual = th_obj(x_th).numpy()
        assert np.allclose(expected, actual)

        actual = th_obj.inverse_transform(torch.from_numpy(expected)).numpy()
        expected = sk_obj.inverse_transform(expected, copy=True)
        assert np.allclose(expected, actual)



from typing import Any, Optional

import torch
import torch.nn as nn


class Kernel(nn.Module):
    def __init__(
        self,
        metric: str,
        gamma: Optional[float] = None,
        coef0: Optional[float] = None,
        degree: Optional[float] = None,
    ):
        super().__init__()
        if metric not in ["linear", "poly", "polynomial", "rbf", "sigmoid"]:
            raise ValueError(f"unsupported kernel metric: {metric}")
        self.metric = metric
        self.has_gamma = gamma is not None
        self.gamma = float(gamma or 0.0)
        self.has_coef0 = coef0 is not None
        self.coef0 = float(coef0 or 0.0)
        self.has_degree = degree is not None
        self.degree = float(degree or 0.0)

    @classmethod
    def wrap(self, estimator: Any) -> "Kernel":
        if not isinstance(estimator.kernel, str):
            raise ValueError(f"kernel must be str, but got {estimator.kernel}")
        return Kernel(
            metric=estimator.kernel,
            gamma=estimator.gamma
            if not hasattr(estimator, "_gamma")
            else estimator._gamma,
            coef0=estimator.coef0,
            degree=estimator.degree,
        )

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        if self.metric == "linear":
            return x @ y.t()
        elif self.metric in ["poly", "polynomial"]:
            return (self._gamma(x) * x @ y.t() + self._coef0()) ** self._degree()
        elif self.metric == "rbf":
            x_norm = ((x ** 2).sum(-1))[:, None]
            y_norm = ((y ** 2).sum(-1))[None]
            dots = x @ y.t()
            dists = (x_norm + y_norm - 2 * dots).clamp_min(0)
            return (-self._gamma(x) * dists).exp()
        elif self.metric == "sigmoid":
            return (self._gamma(x) * x @ y.t() + self._coef0()).tanh()
        raise RuntimeError(f"unsupported kernel: {self.metric}")

    def _gamma(self, x: torch.Tensor) -> float:
        if self.has_gamma:
            return self.gamma
        elif self.metric in ["rbf", "sigmoid", "poly", "polynomial"]:
            return 1.0 / x.shape[1]
        raise RuntimeError(f"unknown default gamma for kernel: {self.metric}")

    def _coef0(self) -> float:
        if self.has_coef0:
            return self.coef0
        elif self.metric in ["sigmoid", "poly", "polynomial"]:
            return 1.0
        raise RuntimeError(f"unknown default coef0 for kernel: {self.metric}")

    def _degree(self) -> float:
        if self.has_degree:
            return self.degree
        elif self.metric in ["poly", "polynomial"]:
            return 3.0
        raise RuntimeError(f"unknown default degree for kernel: {self.metric}")



from copy import deepcopy
from typing import List, Type

import torch
import torch.jit
import torch.nn as nn
import torch.nn.functional as F
from sklearn.linear_model import SGDClassifier


class TorchSGDClassifier(nn.Module):
    def __init__(
        self,
        weights: torch.Tensor,
        biases: torch.Tensor,
        classes: torch.Tensor,
        loss: str,
    ):
        super().__init__()
        self.loss = loss
        self.weights = nn.Parameter(weights)
        self.biases = nn.Parameter(biases)
        self.register_buffer("classes", classes)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [SGDClassifier]

    @classmethod
    def wrap(cls, obj: SGDClassifier) -> "TorchSGDClassifier":
        est = deepcopy(obj)
        est.densify()
        return cls(
            weights=torch.from_numpy(est.coef_),
            biases=torch.from_numpy(est.intercept_),
            classes=torch.from_numpy(est.classes_),
            loss=est.loss,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.

        An alias for self.predict().
        """
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.
        """
        scores = self.decision_function(x)
        if len(scores.shape) == 1:
            indices = (scores > 0).long()
        else:
            indices = scores.argmax(-1)
        return self.classes[indices]

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        outputs = (x @ self.weights.t()) + self.biases
        if outputs.shape[1] == 1:
            return outputs.view(-1)
        return outputs

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_log_proba(x).exp()

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        if self.loss == "log":
            logits = self.decision_function(x)
            if len(logits.shape) == 1:
                return torch.stack(
                    [F.logsigmoid(-logits), F.logsigmoid(logits)], dim=-1
                )
            # This is a one-versus-rest classifier.
            return F.log_softmax(F.logsigmoid(logits), dim=-1)
        else:
            assert False, "probability prediction not supported for loss: " + self.loss



from typing import List, Optional, Type

import torch
import torch.jit
import torch.nn as nn
import torch.nn.functional as F
from sklearn.base import BaseEstimator
from sklearn.ensemble import GradientBoostingClassifier


class _GradientBoostingStage(nn.Module):
    def __init__(self, trees: List[BaseEstimator]):
        super().__init__()
        from .wrap import wrap

        self.trees = nn.ModuleList([wrap(x) for x in trees])

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.stack([tree(x).view(-1) for tree in self.trees], dim=-1)


class TorchGradientBoostingClassifier(nn.Module):
    def __init__(
        self,
        stages: List[_GradientBoostingStage],
        classes: torch.Tensor,
        init: Optional[nn.Module],
        loss: str,
        learning_rate: float,
    ):
        super().__init__()
        assert loss in ["exponential", "deviance"]
        self.stages = nn.ModuleList(stages)
        self.register_buffer("classes", classes)
        self.has_init = init is not None
        if self.has_init:
            self.init = init
        else:
            # Needed to prevent TorchScript from erroring.
            self.init = nn.Identity()

            def dummy_fn(x: torch.Tensor) -> torch.Tensor:
                return x

            self.init.predict_proba = dummy_fn
        self.loss = loss
        self.learning_rate = learning_rate

        dimension = len(stages[0].trees)
        param = next(self.parameters())
        self.register_buffer(
            "zero_out",
            torch.zeros(dimension, dtype=param.dtype, device=param.device),
        )

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [GradientBoostingClassifier]

    @classmethod
    def wrap(cls, obj: GradientBoostingClassifier) -> "TorchGradientBoostingClassifier":
        from .wrap import wrap

        return cls(
            stages=[_GradientBoostingStage(x) for x in obj.estimators_.tolist()],
            classes=torch.from_numpy(obj.classes_),
            init=wrap(obj.init_) if obj.init_ != "zero" else None,
            loss=obj.loss,
            learning_rate=obj.learning_rate,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        if self.loss == "exponential":
            decisions = self.decision_function(x)
            assert len(decisions.shape) == 1
            return self.classes[(decisions >= 0).long()]
        proba = self.predict_proba(x)
        return self.classes[proba.argmax(-1)]

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_log_proba(x).exp()

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        decisions = self.decision_function(x)
        if self.loss == "deviance":
            if len(decisions.shape) == 1:
                return torch.stack(
                    [F.logsigmoid(-decisions), F.logsigmoid(decisions)], dim=-1
                )
            else:
                return F.log_softmax(decisions, dim=-1)
        else:
            return torch.stack(
                [F.logsigmoid(-2 * decisions), F.logsigmoid(2 * decisions)], dim=-1
            )

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        out = self._init_outputs(x)
        for stage in self.stages:
            out = out + stage.forward(x) * self.learning_rate
        if out.shape[1] == 1:
            return out.view(-1)
        return out

    def _init_outputs(self, x: torch.Tensor) -> torch.Tensor:
        if not self.has_init:
            return self.zero_out[None].repeat(len(x), 1)
        eps = 1.1920929e-07  # np.finfo(np.float32).eps
        init_probs = self.init.predict_proba(x).clamp(eps, 1 - eps)
        if self.loss == "exponential":
            assert init_probs.shape[1] == 2
            # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09bcc2eaeba98f7e737aac2ac782f0e5f1/sklearn/ensemble/_gb_losses.py#L978
            prob_pos = init_probs[:, 1]
            return (0.5 * (prob_pos / (1 - prob_pos)).log())[:, None]
        else:
            if init_probs.shape[1] == 2:
                # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09bcc2eaeba98f7e737aac2ac782f0e5f1/sklearn/ensemble/_gb_losses.py#L749
                prob_pos = init_probs[:, 1]
                return ((prob_pos / (1 - prob_pos)).log())[:, None]
            else:
                # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09bcc2eaeba98f7e737aac2ac782f0e5f1/sklearn/ensemble/_gb_losses.py#L866
                return init_probs.log()



from setuptools import setup

setup(
    name="sk2torch",
    version="1.0.0",
    description="Convert scikit-learn models to PyTorch modules",
    long_description="Convert scikit-learn models to PyTorch modules",
    packages=["sk2torch"],
    install_requires=["scikit-learn", "torch"],
    author="Alex Nichol",
    author_email="unixpickle@gmail.com",
    url="https://github.com/unixpickle/sk2torch",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
)



import numpy as np
import pytest
import torch
import torch.jit
from sklearn.kernel_approximation import Nystroem

from .kernel_test import create_test_data
from .nystroem import TorchNystroem


@pytest.mark.parametrize(
    ("kernel", "gamma", "coef0", "degree"),
    [
        ("rbf", None, None, None),
        ("sigmoid", None, None, None),
        ("polynomial", None, None, None),
        ("linear", None, None, None),
        ("rbf", 0.5, None, None),
        ("sigmoid", 0.5, -0.2, None),
        ("polynomial", 0.5, -0.2, 2),
    ],
)
def test_nystroem_defaults(kernel, gamma, coef0, degree):
    x_th, _ = create_test_data()
    x = x_th.numpy()
    sk_obj = Nystroem(
        n_components=5, kernel=kernel, gamma=gamma, coef0=coef0, degree=degree
    )
    sk_obj.fit(x)
    th_obj = torch.jit.script(TorchNystroem.wrap(sk_obj))

    with torch.no_grad():
        expected = sk_obj.transform(x)
        actual = th_obj.transform(x_th).numpy()
        assert (np.abs(actual - expected) < 1e-5).all()



from typing import List, Optional, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.preprocessing import StandardScaler


class TorchStandardScaler(nn.Module):
    def __init__(self, mean: Optional[torch.Tensor], scale: Optional[torch.Tensor]):
        super().__init__()
        self.mean = nn.Parameter(mean if mean is not None else torch.zeros(()))
        self.scale = nn.Parameter(scale if scale is not None else torch.ones(()))

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [StandardScaler]

    @classmethod
    def wrap(cls, obj: StandardScaler) -> "TorchStandardScaler":
        return cls(
            mean=torch.from_numpy(obj.mean_) if obj.with_mean else None,
            scale=torch.from_numpy(obj.scale_) if obj.with_std else None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.transform(x)

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.mean) / self.scale

    @torch.jit.export
    def inverse_transform(self, x: torch.Tensor) -> torch.Tensor:
        return (x * self.scale) + self.mean



from typing import Tuple

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.svm import SVC, LinearSVC, NuSVC

from .svc import TorchLinearSVC, TorchSVC


@pytest.mark.parametrize(
    (
        "kernel",
        "probability",
        "decision_shape",
        "n_classes",
        "break_ties",
        "nu_svc",
        "space_classes",
    ),
    [
        ("rbf", True, "ovr", 4, False, False, False),
        ("rbf", True, "ovr", 4, False, False, True),
        ("rbf", True, "ovr", 2, False, False, False),
        ("rbf", True, "ovr", 4, True, False, False),
        ("rbf", True, "ovr", 2, False, False, False),
        ("rbf", True, "ovo", 4, False, False, False),
        ("rbf", True, "ovo", 2, False, False, False),
        ("rbf", True, "ovo", 2, False, False, False),
        # Different kernels and NuSVC
        ("linear", True, "ovr", 4, False, False, False),
        ("linear", True, "ovr", 2, False, False, False),
        ("linear", True, "ovr", 4, False, True, False),
        ("rbf", True, "ovr", 4, False, True, False),
    ],
)
def test_svc(
    kernel: str,
    probability: bool,
    decision_shape: str,
    n_classes: int,
    break_ties: bool,
    nu_svc: bool,
    space_classes: bool,
):
    xs, ys = quadrant_dataset(n_classes, space_classes)

    test_xs = np.random.random(size=(128, 2)) * 2 - 1
    test_xs_th = torch.from_numpy(test_xs)

    model = (SVC if not nu_svc else NuSVC)(
        kernel=kernel, probability=probability, break_ties=break_ties
    )
    model.fit(xs, ys)
    model.decision_function_shape = decision_shape
    model_th = torch.jit.script(TorchSVC.wrap(model))

    with torch.no_grad():
        expected = model.decision_function(test_xs)
        actual = model_th.decision_function(test_xs_th).numpy()
        assert (np.abs(actual - expected) < 1e-8).all()

        expected = model.predict(test_xs)
        actual = model_th.predict(test_xs_th).numpy()
        assert (actual == expected).all()

        if probability:
            expected = model.predict_proba(test_xs)
            actual = model_th.predict_proba(test_xs_th).numpy()
            # Tolerance of 1e-2 is necessary because sklearn uses
            # a less accurate approximation for the binary case, and
            # more aggressive (non-batched) early stopping for the
            # multiclass case.
            assert (np.abs(actual - expected) < 1e-2).all()

            expected = actual
            actual = model_th.predict_log_proba(test_xs_th).exp().numpy()
            assert (np.abs(actual - expected) < 1e-3).all()


@pytest.mark.parametrize(
    ("fit_intercept", "n_classes", "space_classes"),
    [
        (False, 2, False),
        (True, 2, False),
        (False, 3, False),
        (True, 3, False),
        (True, 4, True),
    ],
)
def test_linear_svc(fit_intercept: bool, n_classes: int, space_classes: int):
    xs, ys = quadrant_dataset(n_classes, space_classes)

    test_xs = np.random.random(size=(128, 2)) * 2 - 1
    test_xs_th = torch.from_numpy(test_xs)

    model = LinearSVC(fit_intercept=fit_intercept, intercept_scaling=2.314)
    model.fit(xs, ys)
    model_th = torch.jit.script(TorchLinearSVC.wrap(model))

    with torch.no_grad():
        expected = model.decision_function(test_xs)
        actual = model_th.decision_function(test_xs_th).numpy()
        assert (np.abs(actual - expected) < 1e-8).all()

        expected = model.predict(test_xs)
        actual = model_th.predict(test_xs_th).numpy()
        assert (actual == expected).all()


def quadrant_dataset(
    n_classes: int, space_classes: bool
) -> Tuple[np.ndarray, np.ndarray]:
    xs = np.random.RandomState(1337).random(size=(500, 2)) * 2 - 1
    ys = np.array([int(x[0] > 0) | (int(x[1] > 0) << 1) for x in xs])

    # Put some empty space around the decision boundary.
    indices = np.max(np.abs(xs) > 0.1, axis=-1)
    xs, ys = xs[indices], ys[indices]

    if n_classes == 2:
        indices = np.logical_or(ys == 0, ys == 1)
        xs, ys = xs[indices], ys[indices]
    elif n_classes == 3:
        indices = np.logical_or(np.logical_or(ys == 0, ys == 1), ys == 2)
        xs, ys = xs[indices], ys[indices]
    else:
        assert n_classes == 4

    if space_classes:
        ys = np.where(ys > n_classes // 2, 1 + (n_classes - ys) + n_classes // 2, ys)

    return xs, ys



import numpy as np
import pytest
import torch
import torch.jit
from sklearn.preprocessing import LabelBinarizer

from .label_binarizer import TorchLabelBinarizer


@pytest.mark.parametrize(
    ("inputs", "neg_label", "pos_label"),
    [
        (np.array([1, 0, 1, 1, 0]), 0, 1),
        (np.array([True, False, True, True, False]), 0, 1),
        (np.array([2, 1, 2, 2, 1]), 0, 1),
        (np.array([1, 0, 1, 1, 0]), 1, 2),
        (np.array([2, 1, 2, 2, 1]), 1, 2),
        (np.array([True, False, True, True, False]), 1, 2),
        (np.array([1, 0, 1, 1, 0]), -1, 0),
        (np.array([1, 0, 1, 1, 0]), -13, 13),
        (np.array([[1, 0], [1, 1], [0, 0]]), 0, 1),
        (np.array([[1, 2], [2, 2], [1, 1]]), 0, 1),
        (np.array([[1, 5], [5, 5], [1, 1]]), 0, 1),
        (np.array([[1, 0], [1, 1], [0, 0]]), 1, 2),
        (np.array([[1, 2], [2, 1], [1, 2]]), 1, 2),
        (np.array([[-1, 5], [5, -1], [-1, 5]]), 1, 2),
        (np.array([[1, 0], [1, 1], [0, 0]]), -1, 0),
        (np.array([[True, False], [True, True], [False, False]]), 0, 1),
        (np.array([[True, False], [True, True], [False, False]]), 1, 2),
        (np.array([[True, False], [True, True], [False, False]]), 3, 4),
        (np.array([[True, False], [True, True], [False, False]]), -1, 0),
        (np.array([0, 1, 2, 3]), 0, 1),
        (np.array([6, 10, 12, 24]), 0, 1),
        (np.array([0, 1, 2, 3]), 1, 2),
        (np.array([6, 10, 12, 24]), 1, 2),
        (np.array([0, 1, 2, 3]), -1, 0),
    ],
)
def test_label_binarizer(inputs: np.ndarray, neg_label: int, pos_label: int):
    sk_obj = LabelBinarizer(neg_label=neg_label, pos_label=pos_label)
    sk_obj.fit(inputs)
    th_obj = torch.jit.script(TorchLabelBinarizer.wrap(sk_obj))

    inputs_th = torch.from_numpy(inputs)

    with torch.no_grad():
        expected = sk_obj.transform(inputs)
        actual = th_obj.transform(inputs_th).numpy()
        assert (expected == actual).all()

        expected = sk_obj.inverse_transform(sk_obj.transform(inputs))
        actual = th_obj.inverse_transform(th_obj.transform(inputs_th)).numpy()
        assert (expected == actual).all()

        xf_shape = sk_obj.transform(inputs).shape
        random_th = torch.rand(100, *xf_shape[1:]).double()
        expected = sk_obj.inverse_transform(random_th.numpy())
        actual = th_obj.inverse_transform(random_th).numpy()
        assert (expected == actual).all()



import itertools

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from .tree import TorchDecisionTreeClassifier, TorchDecisionTreeRegressor


@pytest.mark.parametrize(
    ("target", "max_depth"),
    list(itertools.product(["1d", "1d-nonflat", "2d"], [1, 3, None])),
)
def test_decision_tree_regressor(target: str, max_depth: int):
    xs = np.random.RandomState(1337).normal(size=(1000, 2))
    norms = np.sum(xs ** 2, axis=-1)
    if target == "2d":
        ys = np.stack([norms, np.sqrt(norms)], axis=1)
    else:
        if target == "1d":
            ys = norms
        else:
            ys = norms[:, None]

    sk_obj = DecisionTreeRegressor(max_depth=max_depth)
    sk_obj.fit(xs, ys)
    th_obj = torch.jit.script(TorchDecisionTreeRegressor.wrap(sk_obj))

    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        actual = th_obj.predict(xs_th).numpy()
        expected = sk_obj.predict(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        actual = th_obj.decision_path(xs_th).numpy()
        expected = sk_obj.decision_path(xs)
        assert (actual == expected).all()


@pytest.mark.parametrize(
    ("target", "max_depth"),
    list(itertools.product(["binary", "trinary", "many"], [1, 3, None])),
)
def test_decision_tree_classifier(target: str, max_depth: int):
    xs = np.random.RandomState(1337).normal(size=(1000, 3))
    norms = np.sqrt(np.mean(xs ** 2, axis=-1))
    if target == "binary":
        ys = norms > np.mean(norms)
    elif target == "trinary":
        ys = (
            (norms > np.mean(norms)).astype(np.int32)
            + (norms > np.mean(norms) + 0.1).astype(np.int32)
            + 13
        )
    else:
        # One label with three classes, the other with two.
        y1 = (
            (norms > np.mean(norms)).astype(np.int32)
            + (norms > np.mean(norms) + 0.1).astype(np.int32)
            + 1
        )
        y2 = (norms < np.mean(norms) - 0.1).astype(np.int32) + 3
        ys = np.stack([y1, y2], axis=1)

    sk_obj = DecisionTreeClassifier(max_depth=max_depth)
    sk_obj.fit(xs, ys)
    th_obj = torch.jit.script(TorchDecisionTreeClassifier.wrap(sk_obj))

    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        actual = th_obj.predict(xs_th).numpy()
        expected = sk_obj.predict(xs)
        assert actual.shape == expected.shape
        assert (actual == expected).all()

        actual = th_obj.decision_path(xs_th).numpy()
        expected = sk_obj.decision_path(xs)
        assert (actual == expected).all()

        actual = th_obj.predict_proba(xs_th)
        expected = sk_obj.predict_proba(xs)
        assert isinstance(actual, list) == (len(ys.shape) > 1)
        assert len(actual) == len(expected)
        for a, x in zip(actual, expected):
            a = a.numpy()
            assert a.shape == x.shape
            assert np.allclose(a, x)

        actual = th_obj.predict_log_proba(xs_th)
        with np.errstate(divide="ignore"):  # ignore log of 0
            expected = sk_obj.predict_log_proba(xs)
        assert isinstance(actual, list) == (len(ys.shape) > 1)
        assert len(actual) == len(expected)
        for a, x in zip(actual, expected):
            a = a.numpy()
            assert a.shape == x.shape
            assert np.allclose(np.exp(a), np.exp(x))



from typing import List, Type

import torch
import torch.nn as nn
from sklearn.compose import TransformedTargetRegressor


class TorchTransformedTargetRegressor(nn.Module):
    def __init__(self, regressor: nn.Module, transformer: nn.Module, training_dim: int):
        super().__init__()
        self.regressor = regressor
        self.transformer = transformer
        self.training_dim = training_dim

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [TransformedTargetRegressor]

    @classmethod
    def wrap(cls, obj: TransformedTargetRegressor) -> "TorchTransformedTargetRegressor":
        assert (
            obj.transformer_ is not None
        ), "identity and function transformers not supported"

        from .wrap import wrap

        return TorchTransformedTargetRegressor(
            regressor=wrap(obj.regressor_),
            transformer=wrap(obj.transformer_),
            training_dim=obj._training_dim,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        y = self.regressor(x)
        if len(y.shape) == 1:
            y = y.view(-1, 1)
        y = self.transformer.inverse_transform(y)
        if self.training_dim == 1 and len(y.shape) == 2 and y.shape[1] == 1:
            y = y.squeeze(1)
        return y
